/*
A small tool to scan a DOS system for a possible installed Transputer system.
(c)2011 by Axel Muhr 

As default, port adress 0x150 will be used. This can be changed via parameter.

 Usage: itest [[p][q] r|d|s ] - valid options are:
 /p : set IO Port for [r|d|s] in HEX (0x150 is default)
 /q : Quiet. No screen output, return-codes only.
 /r : resets Transputer at given port-address
 /d : detect link-interface at given port-address
 /s : bruteforce search for linkinterface & Transputer (use with care)
*/

#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <time.h>
#include <errno.h>
#include <string.h>


#define PORT_RD_DATA 0
#define PORT_WR_DATA 1
#define PORT_RD_STAT 2
#define PORT_WR_STAT 3
#define PORT_SS_R    0x10

#define TRUE 1
#define FALSE 0

//int C012_timeout = FALSE;
int Quiet = FALSE;
int PortAddr = 0x150;

int	optind	= 1;	/* index of which argument is next	*/
char   *optarg;		/* pointer to argument of current option */
int	opterr	= 1;	/* allow error message	*/
 
static	char   *letP = NULL;	/* remember next option char's location */
static	char	SW = 0;		/* DOS switch character, either '-' or '/' */

/*
  Parse the command line options, System V style.
 
  Standard option syntax is:
 
    option ::= SW [optLetter]* [argLetter space* argument]
 
  where
    - SW is either '/' or '-'.
    - there is no space before any optLetter or argLetter.
    - opt/arg letters are alphabetic, not punctuation characters.
    - optLetters, if present, must be matched in optionS.
    - argLetters, if present, are found in optionS followed by ':'.
    - argument is any white-space delimited string.  Note that it
      can include the SW character.
    - upper and lower case letters are distinct.
 
  There may be multiple option clusters on a command line, each
  beginning with a SW, but all must appear before any non-option
  arguments (arguments not introduced by SW).  Opt/arg letters may
  be repeated: it is up to the caller to decide if that is an error.
 
  The character SW appearing alone as the last argument is an error.
  The lead-in sequence SWSW ("--" or "//") causes itself and all the
  rest of the line to be ignored (allowing non-options which begin
  with the switch char).

  The string *optionS allows valid opt/arg letters to be recognized.
  argLetters are followed with ':'.  Getopt () returns the value of
  the option character found, or EOF if no more options are in the
  command line.	 If option is an argLetter then the global optarg is
  set to point to the argument string (having skipped any white-space).
 
  The global optind is initially 1 and is always left as the index
  of the next argument of argv[] which getopt has not taken.  Note
  that if "--" or "//" are used then optind is stepped to the next
  argument before getopt() returns EOF.
 
  If an error occurs, that is an SW char precedes an unknown letter,
  then getopt() will return a '?' character and normally prints an
  error message via perror().  If the global variable opterr is set
  to false (zero) before calling getopt() then the error message is
  not printed.
 
  For example, if the MSDOS switch char is '/' (the MSDOS norm) and
 
    *optionS == "A:F:PuU:wXZ:"
 
  then 'P', 'u', 'w', and 'X' are option letters and 'F', 'U', 'Z'
  are followed by arguments.  A valid command line may be:
 
    aCommand  /uPFPi /X /A L someFile
 
  where:
    - 'u' and 'P' will be returned as isolated option letters.
    - 'F' will return with "Pi" as its argument string.
    - 'X' is an isolated option.
    - 'A' will return with "L" as its argument.
    - "someFile" is not an option, and terminates getOpt.  The
      caller may collect remaining arguments using argv pointers.
*/

/* ============================================================================================ */
int getopt( int argc, char** argv, char* opts ) {
/* ============================================================================================ */

    static int sp = 1;
    int c;
    char *cp;

    if ( sp == 1 ) {
        if (    ( optind >= argc )
             || (    ( argv[ optind ][ 0 ] != '-' )
                  && ( argv[ optind ][ 0 ] != '/' ))
             || ( argv[ optind ][ 1 ] == '\0' )) {
            return( EOF );
        } else if ( strcmp( argv[ optind ], "--" ) == 0 ) {
            optind++;
            return( EOF );
        }
    }

    c = argv[ optind ][ sp ];

    if (( c == ':' ) || (( cp = strchr( opts, c )) == NULL )) {
        if( opterr ) {
            (void)fprintf( stderr, "ERROR: illegal option -- %c\n", c );
        }
        if ( argv[ optind ][ ++sp ] == '\0' ) {
            optind++;
            sp = 1;
        }
        return( '?' );
    }

    if( *( ++cp ) == ':' ) {
        if ( argv[ optind ][ sp+1 ] != '\0' ) {
            optarg = &argv[ optind++ ][ sp+1 ];
        } else if ( ++optind >= argc ) {
            if( opterr ) {
                (void)fprintf( stderr, "ERROR: option requires an argument -- %c\n", c );
            }
            sp = 1;
            return( '?' );
        } else {
            optarg = argv[ optind++ ];
        }
        sp = 1;
    } else {
        if ( argv[ optind ][ ++sp ] == '\0' ) {
            sp = 1;
            optind++;
        }
        optarg = NULL;
    }
    return( c );
}

/* ------------------------------------------------------------------------ */
unsigned char RecvByte( void ) {
/* ------------------------------------------------------------------------ */

    while (( inportb( PortAddr + PORT_RD_STAT ) & 1 ) == 0 );
    return inportb( PortAddr + PORT_RD_DATA );	
}

/* ------------------------------------------------------------------------ */
void SendByte( unsigned char b ) {
/* ------------------------------------------------------------------------ */

    while (( inportb( PortAddr + PORT_WR_STAT ) & 1 ) == 0 );
    outportb( PortAddr + PORT_WR_DATA, b );
}

/* ------------------------------------------------------------------------ */
unsigned long int Peek( unsigned long int a ) {
/* ------------------------------------------------------------------------ */

    unsigned long int d;
    unsigned char *p;

    SendByte( 1 );

    p = (unsigned char *)&a;
    SendByte( p[ 0 ] );
    SendByte( p[ 1 ] );
    SendByte( p[ 2 ] );
    SendByte( p[ 3 ] );

    p = (unsigned char *)&d;
    p[ 0 ] = RecvByte();
    p[ 1 ] = RecvByte();
    p[ 2 ] = RecvByte();
    p[ 3 ] = RecvByte();

    return d;
}

/* ------------------------------------------------------------------------ */
void Poke( unsigned long int a, unsigned long int d ) {
/* ------------------------------------------------------------------------ */

    unsigned char *p;

    SendByte( 0 );

    p = (unsigned char *)&a;
    SendByte( p[ 0 ] );
    SendByte( p[ 1 ] );
    SendByte( p[ 2 ] );
    SendByte( p[ 3 ] );

    p = (unsigned char *)&d;
    SendByte( p[ 0 ] );
    SendByte( p[ 1 ] );
    SendByte( p[ 2 ] );
    SendByte( p[ 3 ] );
}

/* ------------------------------------------------------------------------ */
void Reset( void ) {
/* ------------------------------------------------------------------------ */

    outportb( PortAddr + PORT_SS_R, 1 );
    delay( 100 );
    outportb( PortAddr + PORT_SS_R, 0 );
    delay(   1 );
}

void usage() {
	printf("iTest v1.0, Transputer detection tool, (c)2011 by Axel Muhr\n \
 Usage: itest [[p][q] r|d|s ] - valid options are:\n \
 /p : set IO Port for [r|d|s] in HEX (0x150 is default)\n \
 /q : Quiet. No screen output, return-codes only.\n \
 /r : resets Transputer at given port-address\n \
 /d : detect link-interface at given port-address\n \
 /s : bruteforce search for linkinterface & Transputer (use with care)\n \
\n");
	exit(0);
}

unsigned int detect() {
/*  Return Codes of this fuction
10 - No C012 detected
11 - Possible C012 but sent-byte stuck
12 - Possible C012 found
*/	
int RetCode;
	
	Reset();
	outportb( PortAddr + PORT_RD_STAT, 0 ); // Clear Input Interrupt enable
	outportb( PortAddr + PORT_WR_STAT, 0 ); // Clear Input Interrupt enable
	
	// Input status = No data available AND Output status = Ready to send
	if (((inportb( PortAddr + PORT_RD_STAT ) & 1) == 0) && ((inportb( PortAddr + PORT_WR_STAT ) & 1) == 1)) { 
	  outportb( PortAddr + PORT_WR_DATA, 1 ); // Send a byte
	  if (inportb( PortAddr + PORT_RD_STAT ) != 1 ) {
	    RetCode = 12; // Written byte vanished
	  } else {
	    RetCode = 11; // Written byte got stuck
	  }
	} else {
	  RetCode = 10; 
	}
	Reset(); // Keep possible Transputer in sane state
	return (RetCode);
}


unsigned int searchlink() {

/*  Return Codes of this fuction
15 - No C012 detected
16 - Found 16 bit Transputer
32 - Found 32 bit Transputer
23 - Failed to determine Root Transputer type
04 - Found C004
*/
static unsigned char BOOTSTRING[] =
{23,                            /* length of bootstrap  */
 0xB1,                          /* AJW 1, allow for Iptr store    */
 0xD1,                          /* STL 1                          */
 0x24, 0xF2,                    /* MINT                           */
 0x21, 0xFC,                    /* STHF                           */
 0x24, 0xF2,                    /* MINT                           */
 0x21, 0xF8,                    /* STLF                           */
 0xF0,                          /* REV                            */
 0x60, 0x5C,                    /* LDNLP -4                       */
 0x2A, 0x2A, 0x2A, 0x4A,        /* LDC #AAAA                      */
 0xFF,                          /* OUTWORD                        */
 0x21, 0x2F, 0xFF,              /* START                          */
 0x02, 0x00};                   /* C004 read link                 */
 
int i;
int count = 0;
int bytes = 0;
int going = TRUE;
unsigned char buffer[30];


if (detect() == 12) {

//	printf("Debug: sending BOOTSTRING");
	for (i = 0; i < 24; i++) {
		SendByte(BOOTSTRING[i]);
		//printf("%x ", BOOTSTRING[i]);
	}
//	printf(" done.\n");

	    while (going)
	    {
		    time_t StartTime = time(NULL);

		    while (( inportb( PortAddr + PORT_RD_STAT ) & 1 ) == 0 ) {
			if (difftime(time(NULL), StartTime) > 1) {
			  count++;
			  break;
			}
		    }

		    if (count > 0) {
			going = FALSE;
		    } else {
			  buffer[bytes] = inportb( PortAddr + PORT_RD_DATA );
			  bytes++;
		    }
	    }
	    
	    switch (bytes)
	    {
	      case 1:
		return (4); // Found C004
	      case 2:
		if ((buffer[0] == 0xAA) && (buffer[1] == 0xAA))
		{
		  return(16); // Found 16 bit Transputer
		} else
		  return(23); // Failed to determine Root Transputer type
	      case 4:
		if ((buffer[0] == 0xAA) && (buffer[1] == 0xAA))
		{
		  return(32); // Found 32 bit Transputer
		} else
		  return(23); // Failed to determine Root Transputer type
	      default:
		return(23); // Failed to determine Root Transputer type
    	    }
} else {
	return (15);// no link-interface found	
}

}



/* ======================================================================== */
int main( int argc, char* argv[] ) {
/* ======================================================================== */

char c = NULL;
int doSearch, doDetect;

int retval = 0;

    while (c>=0) {
	c = getopt(argc, argv, "p:qsdr?h");
	switch (c) {
		case 'p': PortAddr = strtol(optarg,NULL,0); break;
		case 'q': Quiet = TRUE; break;
		case 's': doSearch = 1; break;
		case 'd': doDetect = 1; break;
		case 'r': Reset(); break;
		case '?':
		case 'h': usage(); break;
	}
    }

  if (doSearch == 1 && doDetect == 1) {
	printf("Error: You can not use the switches -d and -s in the same call.\n");
	exit (-1);
  } else {
	if (doSearch == 1) {
		retval = searchlink();
	}
	if (doDetect == 1) {
		retval = detect();
	}
  }

/*  	Return Codes of "detect()"

10 - No C012 detected
11 - Possible C012 but sent-byte stuck
12 - Possible C012 found

	Return Codes of "searchlink()"

15 - No C012 detected
16 - Found 16 bit Transputer
32 - Found 32 bit Transputer
23 - Failed to determine Root Transputer type
04 - Found C004

*/
	if (Quiet != TRUE) {
	  switch (retval) {
	    case 0 : usage(); break;
	    case 10 : printf("No IMSC011/12 detected.\n"); break;
	    case 11 : printf("Possible C012 detected but sent-byte stuck.\n"); break;
	    case 12 : printf("Possible C012 detected on port %#x.\n", PortAddr); break;
	    case 15 : printf("No IMSC011/12 detected.\n"); break;
	    case 16 : printf("Found 16bit Transputer on port %#x.\n", PortAddr); break;
	    case 32 : printf("Found 32bit Transputer on port %#x.\n", PortAddr); break;
	    case 23 : printf("Failed to determine Root Transputer type"); break;
	    case 4 : printf("Found C004 on port %#x.\n", PortAddr); break;
	  }
	
	}
	
return (retval);
}
